import json
import typing
from abc import abstractmethod

from cloudevents.abstract import AnyCloudEvent
from cloudevents.conversion import to_binary
from cloudevents.http import from_dict
from starlette.background import BackgroundTask
from starlette.responses import JSONResponse

from fastapi_cloudevents.cloudevent import DEFAULT_SOURCE


class _CloudEventResponse(JSONResponse):
    @abstractmethod
    def replace_default_source(self, new_source: str):
        pass


class StructuredCloudEventResponse(_CloudEventResponse):
    """
    Nothing to implement because structured CloudEvents are literally json objects
    """

    @abstractmethod
    def replace_default_source(self, new_source: str):
        result = json.loads(self.body)
        result["source"] = result["source"].replace(DEFAULT_SOURCE, new_source)
        self.body = self.render(result)
        content_length = str(len(self.body))
        headers = dict(self.raw_headers)
        headers[b"content-length"] = content_length.encode("latin-1")
        self.raw_headers = list(headers.items())


class BinaryCloudEventResponse(_CloudEventResponse):
    def __init__(
        self,
        content: typing.Optional[AnyCloudEvent] = None,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTask = None,
    ) -> None:
        super(BinaryCloudEventResponse, self).__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
        )
        self.raw_headers = self._render_headers(content, headers=self.raw_headers)

    def render(self, content: typing.Optional[typing.Any]) -> bytes:
        if content is None:
            return b""
        _, body = to_binary(from_dict(content))
        return body

    @classmethod
    def _render_headers(cls, content: typing.Optional[typing.Any], headers):
        if content is None:
            return headers
        ce_headers, _ = to_binary(from_dict(content))
        ce_headers = [
            (k.encode("utf-8"), v.encode("utf-8")) for k, v in ce_headers.items()
        ]
        result = dict(headers)
        result.update(ce_headers)
        return list(result.items())

    @abstractmethod
    def replace_default_source(self, new_source: str):
        result = dict(self.raw_headers)
        source = result.get(b"ce-source", b"").decode("utf-8")
        source = source.replace(DEFAULT_SOURCE, new_source)
        result[b"ce-source"] = source.encode("utf-8")
        self.raw_headers = list(result.items())



from fastapi_cloudevents.cloudevent import CloudEvent
from fastapi_cloudevents.cloudevent_request import CloudEventRequest
from fastapi_cloudevents.cloudevent_response import (
    BinaryCloudEventResponse, StructuredCloudEventResponse)
from fastapi_cloudevents.cloudevent_route import CloudEventRoute



import cloudevents.pydantic

DEFAULT_SOURCE = "fastapi"


class CloudEvent(cloudevents.pydantic.CloudEvent):
    """
    Same as the official pydantic CloudEvent model, but if no source is given,
    The source will be injected via the CloudEvent response class.
    """

    source: str = DEFAULT_SOURCE



import json
import re

from cloudevents.conversion import to_dict, to_json
from cloudevents.http import from_http
from starlette.requests import Request

_JSON_CONTENT_TYPE = re.compile(r"^.+?/.*\+?json$", flags=re.IGNORECASE)


class CloudEventRequest(Request):
    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            body = await super().body()
            if "ce-specversion" in self.headers:
                event = from_http(dict(self.headers), body)
                if isinstance(event.data, (str, bytes)):
                    datacontenttype = event.get("datacontenttype")
                    if datacontenttype is None or _JSON_CONTENT_TYPE.match(
                        datacontenttype
                    ):
                        try:
                            event.data = json.loads(event.data)
                        except (json.JSONDecodeError, TypeError, UnicodeDecodeError):
                            pass
                body = to_json(event)
                self._json = to_dict(event)
            self._body = body
        return self._body



from typing import Callable

from fastapi.routing import APIRoute
from starlette.requests import Request
from starlette.responses import Response

from fastapi_cloudevents.cloudevent_request import CloudEventRequest
from fastapi_cloudevents.cloudevent_response import _CloudEventResponse
import re

_CE_SOURCE_TAG_PREFIX = re.compile(r"^ce-source:", flags=re.IGNORECASE)


def _is_source_tag(tag: str):
    return _CE_SOURCE_TAG_PREFIX.match(tag)


def _source_tag_to_source(source_tag: str) -> str:
    return _CE_SOURCE_TAG_PREFIX.sub("", source_tag)


def _route_source(route: APIRoute, request: Request) -> str:
    try:
        return next(map(_source_tag_to_source, filter(_is_source_tag, route.tags)))
    except StopIteration:
        return str(request.url)


class CloudEventRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            response = await original_route_handler(
                CloudEventRequest(request.scope, request.receive)
            )
            if isinstance(response, _CloudEventResponse):
                response.replace_default_source(new_source=_route_source(self, request))
            return response

        return custom_route_handler



from pathlib import Path
import setuptools


def _read_requirements(filename):
    return [
        line.strip()
        for line in Path(filename).read_text().splitlines()
        if not line.startswith("#")
    ]


setuptools.setup(
    name="fastapi_cloudevents",
    version="0.1.0b1",
    author="Alexander Tkachev",
    author_email="sasha64sasha@gmail.com",
    description="FastAPI CloudEvents Middleware",
    url="https://github.com/sasha-tkachev/fastapi-cloudevents",
    keywords=[
        "fastapi",
        "cloudevents",
        "k8s",
        "ce",
        "cloud",
        "events",
        "event",
        "middleware",
    ],
    packages=setuptools.find_packages(),
    install_requires=_read_requirements("requirements.txt"),
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 1 - Planning",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    extras_require={
        "tests": [
            "pytest",
            "pytest-cov",
            "coverage",
            "flake8",
        ]
    },
)
